;; pkginfo.el --- Update package-info.in files   -*- mode:emacs-lisp -*-

;; Copyright (C) 2020 Steve Youngs

;; Author:     Steve Youngs <steve@sxemacs.org>
;; Maintainer: Steve Youngs <steve@sxemacs.org>
;; Created:    <2020-03-30>
;; Time-stamp: <Monday Mar 30, 2020 20:43:57 steve>
;; Homepage:   https://git.sxemacs.org/slh
;; Keywords:   packaging, maintenence, convenience

;; This file is part of SLH (Steve's Lisp Hacks)

;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; 3. Neither the name of the author nor the names of any contributors
;;    may be used to endorse or promote products derived from this
;;    software without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
;; IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;; DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
;; BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
;; OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
;; IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;;; Commentary:
;;
;; This is a convenience lib to ease the pain of keeping package-info.in
;; files in check.  Specifically, the `provides' form within those files.
;; If you are not a package maintainer or SXEmacs dev then you do not
;; have any use for this.  Move along, nothing to see here. :-)
;;
;; Getting it working:
;;
;; At a minimum, you need to tweak `pkginfo-dir-toplevel'.  Set it to
;; the toplevel directory of your SXEmacs packages source tree.  It's
;; the one that has the file `XEmacs.rules' in it.  And that's really
;; all you need to set.
;;
;; Usage:
;;
;; Updating a single package's package-info.in
;;    M-x pkginfo-update-provides RET
;;
;; You'll be prompted for the package's directory and after it completes
;; it will pop you into the package-info.in buffer.  If it's all good,
;; just kill the buffer (the file has already been saved).
;;
;; It is also possible to update all standard packages, or all MULE
;; packages, or even all packages (standard + MULE) at once.  See:
;;
;;     #'pkginfo-update-provides-std
;;     #'pkginfo-update-provides-mule
;;     #'pkginfo-update-all-provides
;;
;; In the cases of updating multiple packages at once you will be left
;; in a `buffer-list' buffer that lists just the package-info.in files
;; that were updated.  This is so you can easily check through the
;; changes.
;;
;; Some packages do not get updated because they don't have any
;; provides, and one package doesn't because it shouldn't be touched
;; (it's `leim' and it contains code that is dangerous for SXEmacs
;; because we don't yet have internal unicode support).

;;; Todo:
;;
;;     1) Incorporate this with sy-git.el in some fashion to maybe log
;;        / commit updates.


;;; Code:
;; Both of these are from the 'text-modes' package.
(require 'filladapt)
(require 'tabify)

;; Could be a defcustom, but I seriously don't think it's worth it.
(defvar pkginfo-dir-toplevel
  (file-name-as-directory
   (paths-construct-path
    (list (getenv "HOME") "programming" "SXEmacs" "packages")))
  "Toplevel directory of package source tree.

Value must end with a directory separator.")

(defvar pkginfo-dir-std
  (file-name-as-directory
   (expand-file-name "xemacs-packages" pkginfo-dir-toplevel))
  "Where the normal, non-MULE packages are at.")

(defvar pkginfo-dir-mule
  (file-name-as-directory
   (expand-file-name "mule-packages" pkginfo-dir-toplevel))
  "Where the MULE packages are at.")

(defvar pkginfo-cant-update
  '("general-docs" "psgml-dtds" "sounds-au" "sounds-wav")
  "Non-MULE packages that don't have any provides.")

(defvar pkginfo-cant-update-mule
  '("locale" "leim")
  "MULE packages that shouldn't have their package-info.in updated.")

(defvar pkginfo-lispdir-lisp
  '("bbdb" "gnus" "ess" "jde" "lookup" "mule-ucs" "patcher"
    "riece" "tramp" "vm" "w3" "x-symbol" "xlib" "xslt-process"
    "xwem")
  "A list of packages who's lisp directory is \"lisp\".")

(defvar pkginfo-apel-grep
  (format
   (concat "git grep -E '^%s(product-provide %1$s)?provide ' %2$s/*.el|"
	   "awk -F%3$s '{print $2;}'|"
	   "awk -F%4$s '{print $1;}'|"
	   "tr -d \"')\"|tr -s \"\n\" \" \"")
   (shell-quote-argument "(")
   "%s"
   (shell-quote-argument "'")
   (shell-quote-argument ")"))
  "Shell magic to extract APEL provides.")

(defvar pkginfo-cc-mode-grep
  (format
   (concat "git grep -E '^%s(cc-)?provide ' %s/*.el|"
	   "awk '{print $2;}'|"
	   "tr -d \"')\"|tr -s \"\n\" \" \"")
   (shell-quote-argument "(") "%s")
  "Shell magic to extract cc-mode provides.")

(defun pkginfo-update-provides (pkgdir)
  "Update the package-info.in provides for package in PKGDIR.

Take great care with this and ALWAYS check the results manually.  There
are a couple of packages that are a little, shall we say, quirky, and
their package-info.in will probably need some manual attention.

Quirky Packages:
---------------
    text-modes -- puts a '\"' into the provides
           ede -- has a template file that puts 'ede-proj-%NAME%'
                  into the provides.

There are also packages that will not be updated.  Mostly, they are
packages that do not have any provides.  But there is one, 'leim',
that can not, and must not, be done.  leim contains code that is
dangerous to SXEmacs right now because we do not yet have internal
unicode support.

Packages with NO provides:
-------------------------
    general-docs, psgml-dtds, sounds-au, sounds-wav, locale."
  (interactive
   (list (expand-file-name
	  (read-directory-name "Package Directory: "
			       pkginfo-dir-toplevel nil t
			       "xemacs-packages/"))))
  (let ((scmd (concat "git grep \"^(provide \" %s/*.el|"
		      "awk '{print $2;}'|"
		      "tr -d \"')\"|tr -s \"\n\" \" \""))
	(pinfo (find-file-noselect
		(expand-file-name "package-info.in" pkgdir)))
	(pkg (file-basename pkgdir))
	p1 p2 lispd)
    ;; Can't do these.  They either have no provides, or we shouldn't
    ;; attempt to update them.
    (if (or (member pkg pkginfo-cant-update)
	    (member pkg pkginfo-cant-update-mule))
	(error 'invalid-argument pkg))
    ;; Special case cc-mode (it uses #'cc-provide and #'provide)
    (when (equal pkg "cc-mode")
      (setq scmd pkginfo-cc-mode-grep))
    ;; Special case APEL (allow for its #'product-provide thing)
    (when (equal pkg "apel")
      (setq scmd pkginfo-apel-grep))
    ;; The packages where the elisp isn't in the toplevel pkgdir
    (setq lispd (cond ((member pkg pkginfo-lispdir-lisp) "lisp")
		      ((equal pkg "mew") "mew")
		      ((equal pkg "semantic") "{.,bovine,wisent}")
		      (t ".")))
    (with-current-buffer pinfo
      (goto-char (point-min))
      (setq p1 (search-forward "provides (" nil t))
      (setq p2 (search-forward ")" nil t))
      (delete-region p1 (- p2 1))
      (goto-char p1)
      (insert
       (substring (shell-command-to-string (format scmd lispd))
		  0 -1))
      (filladapt-mode)
      (fill-region p1 (point-at-bol 2))
      (indent-region p1 (point) 8)
      (untabify p1 (point))
      (save-buffer))
    (if (interactive-p)
	(pop-to-buffer pinfo)
      (bury-buffer pinfo))))

(defun pkginfo-update-provides-std ()
  "Update package-info.in provides for all standard \(non-MULE\) pkgs."
  (interactive)
  (let ((pkgs (directory-files
	       pkginfo-dir-std t "[[:alnum:]]$" nil 'subdir)))
    (mapfam
     #'(lambda (pkg)
	 (setq pkgs (delete (expand-file-name pkg pkginfo-dir-std) pkgs)))
     pkginfo-cant-update)
    (unwind-protect
	(mapfam #'pkginfo-update-provides pkgs)
      (when (interactive-p)
	(let ((buf (get-buffer-create "*STD package-infos*")))
	  (list-buffers-internal buf "package-info\\.in.*$")
	  (switch-to-buffer buf))))))

(defun pkginfo-update-provides-mule ()
  "Update package-info.in provides for all MULE packages."
  (interactive)
  (let ((pkgs (directory-files
	       pkginfo-dir-mule t "[[:alnum:]]$" nil 'subdir)))
    (mapfam
     #'(lambda (pkg)
	 (setq pkgs
	       (delete (expand-file-name pkg pkginfo-dir-mule) pkgs)))
     pkginfo-cant-update-mule)
    (unwind-protect
	(mapfam #'pkginfo-update-provides pkgs)
      (when (interactive-p)
	(let ((buf (get-buffer-create "*MULE package-infos*")))
	  (list-buffers-internal buf "package-info\\.in.*$")
	  (switch-to-buffer buf))))))

(defun pkginfo-update-all-provides ()
  "Updates package-info.in provides for all packages, standard + MULE.

Take great care with this and ALWAYS check the results manually.  There
are a couple of packages that are a little, shall we say, quirky, and
their package-info.in will probably need some manual attention.

Quirky Packages:
    text-modes -- puts a '\"' into the provides
           ede -- has a template file that puts 'ede-proj-%NAME%'
                  into the provides.

There are also packages that will not be updated.  Mostly, they are
packages that do not have any provides.  But there is one, 'leim',
that can not, and must not, be done.  leim contains code that is
dangerous to SXEmacs right now because we do not yet have internal
unicode support.

The packages that don't currently have any provides:
    general-docs, psgml-dtds, sounds-au, sounds-wav, locale"
  (interactive)
  (when (yes-or-no-p "Are you ABSOLUTELY SURE you want to do this? ")
    (unwind-protect
	(progn
	  (pkginfo-update-provides-std)
	  (pkginfo-update-provides-mule))
      (let ((buf (get-buffer-create "*All package-infos*")))
	(list-buffers-internal buf "package-info\\.in.*$")
	(switch-to-buffer buf)))))


(provide 'pkginfo)
;;; pkginfo.el ends here
